UNIT Shots;
(* Implements shots. *)

INTERFACE

  CONST
  (* Number of shots. *)
    MAX_SHOTS = 128;

(* (Re)Initializes shots. *)
  PROCEDURE Initialize;
(* Adds a shot.  Returns FALSE if it was no possible (i.e. too much shoots on
  the screen). *)
  FUNCTION Add (CONST aShip, aStraight: BOOLEAN; CONST aX, aY: INTEGER)
    : BOOLEAN;
(* Updates shots. *)
  PROCEDURE Update;
(* Check if given object collide with any shot. *)
  FUNCTION Collide (CONST aShip: BOOLEAN; CONST x, y, w, h: INTEGER): BOOLEAN;
(* Draws shots. *)
  PROCEDURE Draw;

IMPLEMENTATION

  USES
    Aliens, Audio, Display, General, Particles, Player, Sprites,
    allegro5, al5audio;

  TYPE
  (* Defines a shot. *)
    TShot = RECORD
    (* Position. *)
      X, Y: INTEGER;
    (* Motion. *)
      dX, dY: INTEGER;
    (* Current frame. *)
      Frame: INTEGER;
    (* Who shooted. *)
      Ship: BOOLEAN;
    (* Is used. *)
      Used: BOOLEAN;
    END;

  VAR
  (* Shots. *)
    fShot: ARRAY [1..MAX_SHOTS] OF TShot;

(* Initializes. *)
  PROCEDURE Initialize;
  VAR
    Cnt: INTEGER;
  BEGIN
    FOR Cnt := 1 TO MAX_SHOTS DO fShot[Cnt].Used := FALSE
  END;



(* Adds a shot.  Returns FALSE if it was no possible (i.e. too much shoots on
  the screen). *)
  FUNCTION Add (CONST aShip, aStraight: BOOLEAN; CONST aX, aY: INTEGER)
    : BOOLEAN;
  VAR
    Cnt: INTEGER;
    SndEffect: REAL;
  BEGIN
  { Sound effect. }
    IF aShip THEN SndEffect := 1 ELSE SndEffect := FloatBetween (1.5, 1.6);
    al_play_sample (
      SampleShot,
      0.3, 0, SndEffect, ALLEGRO_PLAYMODE_ONCE,
      NIL
    );
  { Look for an unused shot. }
    FOR Cnt := 1 TO MAX_SHOTS DO
      IF NOT fShot[Cnt].Used THEN
      BEGIN
      { Who shoted. }
        fShot[Cnt].Ship := aShip;
        IF aShip THEN
        BEGIN
          fShot[Cnt].x := aX - (SHIP_SHOT_W DIV 2);
          fShot[Cnt].y := aY
        END
        ELSE BEGIN { Alien. }
          fShot[Cnt].x := aX - (ALIEN_SHOT_W DIV 2);
          fShot[Cnt].y := aY - (ALIEN_SHOT_H DIV 2);
          IF aStraight THEN
          BEGIN
            fShot[Cnt].dX := 0;
            fShot[Cnt].dY := 2
          END
          ELSE BEGIN
            fShot[Cnt].dX := IntegerBetween (-2, 2);
            fShot[Cnt].dY := IntegerBetween (-2, 2)
          END;
        { If the shot has no speed, don't bother. }
          IF (fShot[Cnt].dX OR fShot[Cnt].dY) = 0 THEN EXIT (TRUE);
        END;
        fShot[Cnt].Frame := 0;
        fShot[Cnt].Used := TRUE;
      { Shoted. }
        EXIT (TRUE)
      END;
    EXIT (FALSE)
  END;



(* Updates shots. *)
  PROCEDURE Update;
  VAR
    Ndx: INTEGER;
  BEGIN
    FOR Ndx := 1 TO MAX_SHOTS DO
      IF fShot[Ndx].Used THEN
      BEGIN
        IF fShot[Ndx].Ship THEN
        BEGIN
          DEC (fShot[Ndx].y, 5);

          IF fShot[Ndx].y < -SHIP_SHOT_H THEN
            fShot[Ndx].Used := FALSE
        END
        ELSE BEGIN { Alien. }
          INC (fShot[Ndx].x, fShot[Ndx].dX);
          INC (fShot[Ndx].y, fShot[Ndx].dY);

          IF (-ALIEN_SHOT_W > fShot[Ndx].x) OR (fShot[Ndx].x > BUFFER_W)
          OR (-ALIEN_SHOT_H > fShot[Ndx].y) OR (fShot[Ndx].y > BUFFER_H)
          THEN
            fShot[Ndx].Used := FALSE
        END;
        IF fShot[Ndx].Used THEN INC (fShot[Ndx].Frame)
      END
  END;



(* Check collisions with shots. *)
  FUNCTION Collide (CONST aShip: BOOLEAN; CONST x, y, w, h: INTEGER): BOOLEAN;
  VAR
    Ndx, sw, sh: INTEGER;
  BEGIN
  { Shoot size. }
    IF aShip THEN
    BEGIN
      sw := ALIEN_SHOT_W; sh := ALIEN_SHOT_H
    END
    ELSE BEGIN
      sw := SHIP_SHOT_W; sh := SHIP_SHOT_H
    END;
  { Check collisions. }
    FOR Ndx := 1 TO MAX_SHOTS DO
      IF fShot[Ndx].Used
    { Don't collide with one's own shots. }
      AND (fShot[Ndx].Ship <> aShip)
      THEN BEGIN
        IF Sprites.Collide (
           x, y, x + w, y + h,
           fShot[Ndx].x, fShot[Ndx].y, fShot[Ndx].x + sw, fShot[Ndx].y + sh
        ) THEN
        BEGIN
          Particles.Add (
            TRUE,
            fShot[Ndx].x + (sw DIV 2), fShot[Ndx].y + (sh DIV 2)
          );
          fShot[Ndx].Used := FALSE;
          EXIT (TRUE)
        END
      END;
    EXIT (FALSE)
  END;



(* Draws shots. *)
  PROCEDURE Draw;
  VAR
    Ndx, lFrame: INTEGER;
    Tint: ALLEGRO_COLOR;
  BEGIN
    FOR Ndx := 1 TO MAX_SHOTS DO
      IF fShot[Ndx].Used THEN
      BEGIN
        lFrame := (fShot[Ndx].Frame DIV 2) MOD 2;

        IF fShot[Ndx].Ship THEN
          al_draw_bitmap (
            Sprites.ShipShot[lFrame],
            fShot[Ndx].x, fShot[Ndx].y,
            0
          )
        ELSE BEGIN { Alien. }
          IF lFrame <> 0 THEN
            Tint := al_map_rgb_f (1, 1, 1)
          ELSE
            Tint := al_map_rgb_f (0.5, 0.5, 0.5);
          al_draw_tinted_bitmap (
            Sprites.AlienShot,
            Tint, fShot[Ndx].x, fShot[Ndx].y,
            0
          )
        END
      END
  END;

END.
